@@ -45,6 +45,14 @@ class RedisClient
     def default_driver=(name)
       @default_driver = driver(name)
     end
+
+    def now
+      Process.clock_gettime(Process::CLOCK_MONOTONIC)
+    end
+
+    def now_ms
+      Process.clock_gettime(Process::CLOCK_MONOTONIC, :float_millisecond)
+    end
   end
 
   register_driver :ruby do
@@ -91,13 +99,49 @@ class RedisClient
     end
   end
 
+  module Retriable
+    def _set_retry_attempt(retry_attempt)
+      @retry_attempt = retry_attempt
+    end
+
+    def retry_attempt
+      @retry_attempt || 0
+    end
+
+    def retriable?
+      !!@retry_attempt
+    end
+
+    def final?
+      !@retry_attempt
+    end
+  end
+
+  module Final
+    def _set_retry_attempt(_retry_attempt)
+    end
+
+    def retry_attempt
+      0
+    end
+
+    def retriable?
+      false
+    end
+
+    def final?
+      true
+    end
+  end
+
   class Error < StandardError
     include HasConfig
+    include Retriable
 
     def self.with_config(message, config = nil)
-      new(message).tap do |error|
-        error._set_config(config)
-      end
+      error = new(message)
+      error._set_config(config)
+      error
     end
   end
 
@@ -122,8 +166,19 @@ class RedisClient
     end
   end
 
+  module HasCode
+    attr_reader :code
+
+    def initialize(message = nil, code = nil)
+      super(message)
+      @code = code
+    end
+  end
+
   class CommandError < Error
     include HasCommand
+    include HasCode
+    include Final
 
     class << self
       def parse(error_message)
@@ -136,7 +191,7 @@ class RedisClient
         end
         code ||= error_message.split(' ', 2).first
         klass = ERRORS.fetch(code, self)
-        klass.new(error_message.strip)
+        klass.new(error_message.strip, code.freeze)
       end
     end
   end
@@ -145,12 +200,15 @@ class RedisClient
   PermissionError = Class.new(CommandError)
   WrongTypeError = Class.new(CommandError)
   OutOfMemoryError = Class.new(CommandError)
+  NoScriptError = Class.new(CommandError)
 
   ReadOnlyError = Class.new(ConnectionError)
   ReadOnlyError.include(HasCommand)
+  ReadOnlyError.include(HasCode)
 
   MasterDownError = Class.new(ConnectionError)
   MasterDownError.include(HasCommand)
+  MasterDownError.include(HasCode)
 
   CommandError::ERRORS = {
     "WRONGPASS" => AuthenticationError,
@@ -159,6 +217,7 @@ class RedisClient
     "MASTERDOWN" => MasterDownError,
     "WRONGTYPE" => WrongTypeError,
     "OOM" => OutOfMemoryError,
+    "NOSCRIPT" => NoScriptError,
   }.freeze
 
   class << self
@@ -190,6 +249,7 @@ class RedisClient
     @middlewares = config.middlewares_stack.new(self)
     @raw_connection = nil
     @disable_reconnection = false
+    @retry_attempt = nil
   end
 
   def inspect
@@ -684,6 +744,7 @@ class RedisClient
     close if !config.inherit_socket && @pid != PIDCache.pid
 
     if @disable_reconnection
+      @raw_connection.retry_attempt = nil
       if block_given?
         yield @raw_connection
       else
@@ -694,6 +755,7 @@ class RedisClient
       connection = nil
       preferred_error = nil
       begin
+        @retry_attempt = config.retriable?(tries) ? tries : nil
         connection = raw_connection
         if block_given?
           yield connection
@@ -702,9 +764,14 @@ class RedisClient
         end
       rescue ConnectionError, ProtocolError => error
         preferred_error ||= error
-        preferred_error = error unless error.is_a?(CircuitBreaker::OpenCircuitError)
         close
 
+        if error.is_a?(CircuitBreaker::OpenCircuitError)
+          raise preferred_error
+        else
+          preferred_error = error
+        end
+
         if !@disable_reconnection && config.retry_connecting?(tries, error)
           tries += 1
           retry
@@ -717,6 +784,7 @@ class RedisClient
       connection = ensure_connected
       begin
         @disable_reconnection = true
+        @raw_connection.retry_attempt = nil
         yield connection
       rescue ConnectionError, ProtocolError
         close
@@ -731,13 +799,14 @@ class RedisClient
     if @raw_connection.nil? || !@raw_connection.revalidate
       connect
     end
+    @raw_connection.retry_attempt = @retry_attempt
     @raw_connection
   end
 
   def connect
     @pid = PIDCache.pid
 
-    if @raw_connection
+    if @raw_connection&.revalidate
       @middlewares.connect(config) do
         @raw_connection.reconnect
       end
@@ -751,6 +820,7 @@ class RedisClient
         )
       end
     end
+    @raw_connection.retry_attempt = @retry_attempt
 
     prelude = config.connection_prelude.dup
 
