@@ -12,8 +12,8 @@ class RedisClient
     DEFAULT_DB = 0
 
     module Common
-      attr_reader :db, :password, :id, :ssl, :ssl_params, :command_builder, :inherit_socket,
-        :connect_timeout, :read_timeout, :write_timeout, :driver, :connection_prelude, :protocol,
+      attr_reader :db, :id, :ssl, :ssl_params, :command_builder, :inherit_socket,
+        :connect_timeout, :read_timeout, :write_timeout, :driver, :protocol,
         :middlewares_stack, :custom, :circuit_breaker
 
       alias_method :ssl?, :ssl
@@ -40,7 +40,7 @@ class RedisClient
         circuit_breaker: nil
       )
         @username = username
-        @password = password
+        @password = password && !password.respond_to?(:call) ? ->(_) { password } : password
         @db = begin
           Integer(db || DEFAULT_DB)
         rescue ArgumentError
@@ -70,7 +70,6 @@ class RedisClient
 
         reconnect_attempts = Array.new(reconnect_attempts, 0).freeze if reconnect_attempts.is_a?(Integer)
         @reconnect_attempts = reconnect_attempts
-        @connection_prelude = build_connection_prelude
 
         circuit_breaker = CircuitBreaker.new(**circuit_breaker) if circuit_breaker.is_a?(Hash)
         if @circuit_breaker = circuit_breaker
@@ -87,6 +86,39 @@ class RedisClient
         @middlewares_stack = middlewares_stack
       end
 
+      def connection_prelude
+        prelude = []
+        pass = password
+        if protocol == 3
+          prelude << if pass
+            ["HELLO", "3", "AUTH", username, pass]
+          else
+            ["HELLO", "3"]
+          end
+        elsif pass
+          prelude << if @username && !@username.empty?
+            ["AUTH", @username, pass]
+          else
+            ["AUTH", pass]
+          end
+        end
+
+        if @db && @db != 0
+          prelude << ["SELECT", @db.to_s]
+        end
+
+        # Deep freeze all the strings and commands
+        prelude.map! do |commands|
+          commands = commands.map { |str| str.frozen? ? str : str.dup.freeze }
+          commands.freeze
+        end
+        prelude.freeze
+      end
+
+      def password
+        @password&.call(username)
+      end
+
       def username
         @username || DEFAULT_USERNAME
       end
@@ -108,6 +140,10 @@ class RedisClient
         @client_implementation.new(self, **kwargs)
       end
 
+      def retriable?(attempt)
+        @reconnect_attempts && @reconnect_attempts[attempt]
+      end
+
       def retry_connecting?(attempt, _error)
         if @reconnect_attempts
           if (pause = @reconnect_attempts[attempt])
@@ -146,41 +182,11 @@ class RedisClient
         end
         url
       end
-
-      private
-
-      def build_connection_prelude
-        prelude = []
-        if protocol == 3
-          prelude << if @password
-            ["HELLO", "3", "AUTH", @username || DEFAULT_USERNAME, @password]
-          else
-            ["HELLO", "3"]
-          end
-        elsif @password
-          prelude << if @username && !@username.empty?
-            ["AUTH", @username, @password]
-          else
-            ["AUTH", @password]
-          end
-        end
-
-        if @db && @db != 0
-          prelude << ["SELECT", @db.to_s]
-        end
-
-        # Deep freeze all the strings and commands
-        prelude.map! do |commands|
-          commands = commands.map { |str| str.frozen? ? str : str.dup.freeze }
-          commands.freeze
-        end
-        prelude.freeze
-      end
     end
 
     include Common
 
-    attr_reader :host, :port, :path
+    attr_reader :host, :port, :path, :server_key
 
     def initialize(
       url: nil,
@@ -189,14 +195,15 @@ class RedisClient
       path: nil,
       username: nil,
       password: nil,
+      db: nil,
       **kwargs
     )
       if url
         url_config = URLConfig.new(url)
         kwargs = {
           ssl: url_config.ssl?,
-          db: url_config.db,
         }.compact.merge(kwargs)
+        db ||= url_config.db
         host ||= url_config.host
         port ||= url_config.port
         path ||= url_config.path
@@ -204,7 +211,7 @@ class RedisClient
         password ||= url_config.password
       end
 
-      super(username: username, password: password, **kwargs)
+      super(username: username, password: password, db: db, **kwargs)
 
       if @path = path
         @host = nil
@@ -213,6 +220,8 @@ class RedisClient
         @host = host || DEFAULT_HOST
         @port = Integer(port || DEFAULT_PORT)
       end
+
+      @server_key = [@path, @host, @port].freeze
     end
   end
 end
