@@ -38,9 +38,14 @@ class RedisClient
       end
 
       @to_list_of_hash = @to_hash = nil
+      password = if sentinel_password && !sentinel_password.respond_to?(:call)
+        ->(_) { sentinel_password }
+      else
+        sentinel_password
+      end
       @extra_config = {
         username: sentinel_username,
-        password: sentinel_password,
+        password: password,
         db: nil,
       }
       if client_config[:protocol] == 2
@@ -77,6 +82,10 @@ class RedisClient
       end
     end
 
+    def server_key
+      config.server_key
+    end
+
     def host
       config.host
     end
